@@ -33,12 +33,25 @@
 using System;
 using System.IO;
 using Google.Protobuf.TestProtos;
+using Proto2 = Google.Protobuf.TestProtos.Proto2;
 using NUnit.Framework;
 
 namespace Google.Protobuf
 {
     public class UnknownFieldSetTest
     {
+        public class Data
+        {
+            public static System.Collections.IEnumerable Messages
+            {
+                get
+                {
+                    yield return SampleMessages.CreateFullTestAllTypesProto2();
+                    yield return SampleMessages.CreateFullTestAllTypes();
+                }
+            }
+        }
+
         [Test]
         public void EmptyUnknownFieldSet()
         {
@@ -60,24 +73,23 @@ namespace Google.Protobuf
         }
 
         [Test]
-        public void TestMergeCodedInput()
+        [TestCaseSource(typeof(Data), "Messages")]
+        public void TestMergeCodedInput(IMessage message)
         {
-            var message = SampleMessages.CreateFullTestAllTypes();
             var emptyMessage = new TestEmptyMessage();
             emptyMessage.MergeFrom(message.ToByteArray());
             Assert.AreEqual(message.CalculateSize(), emptyMessage.CalculateSize());
             Assert.AreEqual(message.ToByteArray(), emptyMessage.ToByteArray());
 
-            var newMessage = new TestAllTypes();
-            newMessage.MergeFrom(emptyMessage.ToByteArray());
+            var newMessage = message.Descriptor.Parser.ParseFrom(emptyMessage.ToByteArray());
             Assert.AreEqual(message, newMessage);
             Assert.AreEqual(message.CalculateSize(), newMessage.CalculateSize());
         }
 
         [Test]
-        public void TestMergeMessage()
+        [TestCaseSource(typeof(Data), "Messages")]
+        public void TestMergeMessage(IMessage message)
         {
-            var message = SampleMessages.CreateFullTestAllTypes();
             var emptyMessage = new TestEmptyMessage();
             var otherEmptyMessage = new TestEmptyMessage();
             emptyMessage.MergeFrom(message.ToByteArray());
@@ -88,9 +100,9 @@ namespace Google.Protobuf
         }
 
         [Test]
-        public void TestEquals()
+        [TestCaseSource(typeof(Data), "Messages")]
+        public void TestEquals(IMessage message)
         {
-            var message = SampleMessages.CreateFullTestAllTypes();
             var emptyMessage = new TestEmptyMessage();
             var otherEmptyMessage = new TestEmptyMessage();
             Assert.AreEqual(emptyMessage, otherEmptyMessage);
@@ -101,9 +113,9 @@ namespace Google.Protobuf
         }
 
         [Test]
-        public void TestHashCode()
+        [TestCaseSource(typeof(Data), "Messages")]
+        public void TestHashCode(IMessage message)
         {
-            var message = SampleMessages.CreateFullTestAllTypes();
             var emptyMessage = new TestEmptyMessage();
             int hashCode = emptyMessage.GetHashCode();
             emptyMessage.MergeFrom(message.ToByteArray());
@@ -111,7 +123,8 @@ namespace Google.Protobuf
         }
 
         [Test]
-        public void TestClone()
+        [TestCaseSource(typeof(Data), "Messages")]
+        public void TestClone(IMessage message)
         {
             var emptyMessage = new TestEmptyMessage();
             var otherEmptyMessage = new TestEmptyMessage();
@@ -119,7 +132,6 @@ namespace Google.Protobuf
             Assert.AreEqual(emptyMessage.CalculateSize(), otherEmptyMessage.CalculateSize());
             Assert.AreEqual(emptyMessage.ToByteArray(), otherEmptyMessage.ToByteArray());
 
-            var message = SampleMessages.CreateFullTestAllTypes();
             emptyMessage.MergeFrom(message.ToByteArray());
             otherEmptyMessage = emptyMessage.Clone();
             Assert.AreEqual(message.CalculateSize(), otherEmptyMessage.CalculateSize());
@@ -127,9 +139,9 @@ namespace Google.Protobuf
         }
 
         [Test]
-        public void TestDiscardUnknownFields()
+        [TestCaseSource(typeof(Data), "Messages")]
+        public void TestDiscardUnknownFields(IMessage message)
         {
-            var message = SampleMessages.CreateFullTestAllTypes();
             var goldenEmptyMessage = new TestEmptyMessage();
             byte[] data = message.ToByteArray();
             int fullSize = message.CalculateSize();
@@ -149,10 +161,10 @@ namespace Google.Protobuf
             MessageParser discardingParser2 = retainingParser2.WithDiscardUnknownFields(true);
 
             // Test parse from byte[]
-            assertFull(retainingParser1.ParseFrom(data));
-            assertFull(retainingParser2.ParseFrom(data));
-            assertEmpty(discardingParser1.ParseFrom(data));
-            assertEmpty(discardingParser2.ParseFrom(data));
+            MessageParsingHelpers.AssertReadingMessage(retainingParser1, data, m => assertFull(m));
+            MessageParsingHelpers.AssertReadingMessage(retainingParser2, data, m => assertFull(m));
+            MessageParsingHelpers.AssertReadingMessage(discardingParser1, data, m => assertEmpty(m));
+            MessageParsingHelpers.AssertReadingMessage(discardingParser2, data, m => assertEmpty(m));
 
             // Test parse from byte[] with offset
             assertFull(retainingParser1.ParseFrom(data, 0, data.Length));
