@@ -30,6 +30,7 @@
 
 package com.google.protobuf;
 
+import protobuf_unittest.Engine;
 import protobuf_unittest.Vehicle;
 import protobuf_unittest.Wheel;
 import java.util.ArrayList;
@@ -37,8 +38,8 @@ import java.util.List;
 import junit.framework.TestCase;
 
 /**
- * Test cases that exercise end-to-end use cases involving
- * {@link SingleFieldBuilder} and {@link RepeatedFieldBuilder}.
+ * Test cases that exercise end-to-end use cases involving {@link SingleFieldBuilder} and {@link
+ * RepeatedFieldBuilder}.
  *
  * @author jonp@google.com (Jon Perlow)
  */
@@ -46,20 +47,11 @@ public class NestedBuildersTest extends TestCase {
 
   public void testMessagesAndBuilders() {
     Vehicle.Builder vehicleBuilder = Vehicle.newBuilder();
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(4)
-        .setWidth(1);
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(4)
-        .setWidth(2);
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(4)
-        .setWidth(3);
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(4)
-        .setWidth(4);
-    vehicleBuilder.getEngineBuilder()
-        .setLiters(10);
+    vehicleBuilder.addWheelBuilder().setRadius(4).setWidth(1);
+    vehicleBuilder.addWheelBuilder().setRadius(4).setWidth(2);
+    vehicleBuilder.addWheelBuilder().setRadius(4).setWidth(3);
+    vehicleBuilder.addWheelBuilder().setRadius(4).setWidth(4);
+    vehicleBuilder.getEngineBuilder().setLiters(10);
 
     Vehicle vehicle = vehicleBuilder.build();
     assertEquals(4, vehicle.getWheelCount());
@@ -71,11 +63,9 @@ public class NestedBuildersTest extends TestCase {
     assertEquals(10, vehicle.getEngine().getLiters());
 
     for (int i = 0; i < 4; i++) {
-      vehicleBuilder.getWheelBuilder(i)
-          .setRadius(5)
-          .setWidth(i + 10);
+      vehicleBuilder.getWheelBuilder(i).setRadius(5).setWidth(i + 10);
     }
-    vehicleBuilder.getEngineBuilder().setLiters(20);
+    Engine.Builder engineBuilder = vehicleBuilder.getEngineBuilder().setLiters(20);
 
     vehicle = vehicleBuilder.build();
     for (int i = 0; i < 4; i++) {
@@ -85,22 +75,17 @@ public class NestedBuildersTest extends TestCase {
     }
     assertEquals(20, vehicle.getEngine().getLiters());
     assertTrue(vehicle.hasEngine());
+
+    engineBuilder.setLiters(50);
+    assertEquals(50, vehicleBuilder.getEngine().getLiters());
   }
 
   public void testMessagesAreCached() {
     Vehicle.Builder vehicleBuilder = Vehicle.newBuilder();
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(1)
-        .setWidth(2);
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(3)
-        .setWidth(4);
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(5)
-        .setWidth(6);
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(7)
-        .setWidth(8);
+    vehicleBuilder.addWheelBuilder().setRadius(1).setWidth(2);
+    vehicleBuilder.addWheelBuilder().setRadius(3).setWidth(4);
+    vehicleBuilder.addWheelBuilder().setRadius(5).setWidth(6);
+    vehicleBuilder.addWheelBuilder().setRadius(7).setWidth(8);
 
     // Make sure messages are cached.
     List<Wheel> wheels = new ArrayList<Wheel>(vehicleBuilder.getWheelList());
@@ -117,8 +102,7 @@ public class NestedBuildersTest extends TestCase {
     }
 
     // Change just one
-    vehicleBuilder.getWheelBuilder(3)
-        .setRadius(20).setWidth(20);
+    vehicleBuilder.getWheelBuilder(3).setRadius(20).setWidth(20);
 
     // Now get wheels and check that only that one changed
     for (int i = 0; i < wheels.size(); i++) {
@@ -132,12 +116,8 @@ public class NestedBuildersTest extends TestCase {
 
   public void testRemove_WithNestedBuilders() {
     Vehicle.Builder vehicleBuilder = Vehicle.newBuilder();
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(1)
-        .setWidth(1);
-    vehicleBuilder.addWheelBuilder()
-        .setRadius(2)
-        .setWidth(2);
+    vehicleBuilder.addWheelBuilder().setRadius(1).setWidth(1);
+    vehicleBuilder.addWheelBuilder().setRadius(2).setWidth(2);
     vehicleBuilder.removeWheel(0);
 
     assertEquals(1, vehicleBuilder.getWheelCount());
@@ -146,12 +126,8 @@ public class NestedBuildersTest extends TestCase {
 
   public void testRemove_WithNestedMessages() {
     Vehicle.Builder vehicleBuilder = Vehicle.newBuilder();
-    vehicleBuilder.addWheel(Wheel.newBuilder()
-        .setRadius(1)
-        .setWidth(1));
-    vehicleBuilder.addWheel(Wheel.newBuilder()
-        .setRadius(2)
-        .setWidth(2));
+    vehicleBuilder.addWheel(Wheel.newBuilder().setRadius(1).setWidth(1));
+    vehicleBuilder.addWheel(Wheel.newBuilder().setRadius(2).setWidth(2));
     vehicleBuilder.removeWheel(0);
 
     assertEquals(1, vehicleBuilder.getWheelCount());
@@ -159,14 +135,13 @@ public class NestedBuildersTest extends TestCase {
   }
 
   public void testMerge() {
-    Vehicle vehicle1 = Vehicle.newBuilder()
-        .addWheel(Wheel.newBuilder().setRadius(1).build())
-        .addWheel(Wheel.newBuilder().setRadius(2).build())
-        .build();
+    Vehicle vehicle1 =
+        Vehicle.newBuilder()
+            .addWheel(Wheel.newBuilder().setRadius(1).build())
+            .addWheel(Wheel.newBuilder().setRadius(2).build())
+            .build();
 
-    Vehicle vehicle2 = Vehicle.newBuilder()
-        .mergeFrom(vehicle1)
-        .build();
+    Vehicle vehicle2 = Vehicle.newBuilder().mergeFrom(vehicle1).build();
     // List should be the same -- no allocation
     assertSame(vehicle1.getWheelList(), vehicle2.getWheelList());
 
