@@ -5,13 +5,23 @@ set -ex
 # change to repo root
 pushd $(dirname $0)/../../../..
 
+# Create stage dir
+ORIGINAL_DIR=`pwd`
+pushd ..
+cp -R $ORIGINAL_DIR stage
+export STAGE_DIR="`pwd`/stage"
+popd
+
 export REPO_DIR=protobuf
 export BUILD_VERSION=`grep -i "version" python/google/protobuf/__init__.py | grep -o "'.*'" | tr -d "'"`
-export BUILD_COMMIT=v$BUILD_VERSION
+
+export BUILD_COMMIT=`git rev-parse HEAD`
 export PLAT=x86_64
 export UNICODE_WIDTH=32
 export MACOSX_DEPLOYMENT_TARGET=10.9
 
+rm -rf artifacts/
+rm -rf multibuild/
 mkdir artifacts
 export ARTIFACT_DIR=$(pwd)/artifacts
 
@@ -23,8 +33,8 @@ build_artifact_version() {
 
   # Clean up env
   rm -rf venv
-  sudo rm -rf protobuf
-  git clone https://github.com/google/protobuf.git
+  sudo rm -rf $REPO_DIR
+  cp -R $STAGE_DIR $REPO_DIR
 
   source multibuild/common_utils.sh
   source multibuild/travis_steps.sh
@@ -40,7 +50,7 @@ build_artifact_version() {
 }
 
 build_artifact_version 2.7
-build_artifact_version 3.4
 build_artifact_version 3.5
 build_artifact_version 3.6
 build_artifact_version 3.7
+build_artifact_version 3.8
