@@ -25,6 +25,62 @@ case "${ACTION}" in
     ;;
 esac
 
+# -----------------------------------------------------------------------------
+# Reusing a bunch of the protos from the protocolbuffers/protobuf tree, this
+# can include some extras as there is no harm in ensuring work for C++
+# generation.
+
+CORE_PROTO_FILES=(
+  src/google/protobuf/any_test.proto
+  src/google/protobuf/unittest_arena.proto
+  src/google/protobuf/unittest_custom_options.proto
+  src/google/protobuf/unittest_enormous_descriptor.proto
+  src/google/protobuf/unittest_embed_optimize_for.proto
+  src/google/protobuf/unittest_empty.proto
+  src/google/protobuf/unittest_import.proto
+  src/google/protobuf/unittest_import_lite.proto
+  src/google/protobuf/unittest_lite.proto
+  src/google/protobuf/unittest_mset.proto
+  src/google/protobuf/unittest_mset_wire_format.proto
+  src/google/protobuf/unittest_no_generic_services.proto
+  src/google/protobuf/unittest_optimize_for.proto
+  src/google/protobuf/unittest.proto
+  src/google/protobuf/unittest_import_public.proto
+  src/google/protobuf/unittest_import_public_lite.proto
+  src/google/protobuf/unittest_drop_unknown_fields.proto
+  src/google/protobuf/unittest_preserve_unknown_enum.proto
+  src/google/protobuf/map_lite_unittest.proto
+  src/google/protobuf/map_proto2_unittest.proto
+  src/google/protobuf/map_unittest.proto
+  # The unittest_custom_options.proto extends the messages in descriptor.proto
+  # so we build it in to test extending in general. The library doesn't provide
+  # a descriptor as it doesn't use the classes/enums.
+  src/google/protobuf/descriptor.proto
+)
+
+# -----------------------------------------------------------------------------
+# The objc unittest specific proto files.
+
+OBJC_TEST_PROTO_FILES=(
+  objectivec/Tests/unittest_cycle.proto
+  objectivec/Tests/unittest_deprecated.proto
+  objectivec/Tests/unittest_deprecated_file.proto
+  objectivec/Tests/unittest_extension_chain_a.proto
+  objectivec/Tests/unittest_extension_chain_b.proto
+  objectivec/Tests/unittest_extension_chain_c.proto
+  objectivec/Tests/unittest_extension_chain_d.proto
+  objectivec/Tests/unittest_extension_chain_e.proto
+  objectivec/Tests/unittest_extension_chain_f.proto
+  objectivec/Tests/unittest_extension_chain_g.proto
+  objectivec/Tests/unittest_objc.proto
+  objectivec/Tests/unittest_objc_startup.proto
+  objectivec/Tests/unittest_objc_options.proto
+  objectivec/Tests/unittest_runtime_proto2.proto
+  objectivec/Tests/unittest_runtime_proto3.proto
+)
+
+OBJC_EXTENSIONS=( .pbobjc.h .pbobjc.m )
+
 # -----------------------------------------------------------------------------
 # Ensure the output dir exists
 mkdir -p "${OUTPUT_DIR}/google/protobuf"
@@ -37,11 +93,27 @@ cd "${SRCROOT}/.."
   die "Could not find the protoc binary; make sure you have built it (objectivec/DevTools/full_mac_build.sh -h)."
 
 # -----------------------------------------------------------------------------
-# See the compiler or proto files have changed.
 RUN_PROTOC=no
-if [[ ! -d "${OUTPUT_DIR}" ]] ; then
-  RUN_PROTOC=yes
-else
+
+# Check to if all the output files exist (incase a new one got added).
+
+for PROTO_FILE in "${CORE_PROTO_FILES[@]}" "${OBJC_TEST_PROTO_FILES[@]}"; do
+  DIR=${PROTO_FILE%/*}
+  BASE_NAME=${PROTO_FILE##*/}
+  # Drop the extension
+  BASE_NAME=${BASE_NAME%.*}
+  OBJC_NAME=$(echo "${BASE_NAME}" | awk -F _ '{for(i=1; i<=NF; i++) printf "%s", toupper(substr($i,1,1)) substr($i,2);}')
+
+  for EXT in "${OBJC_EXTENSIONS[@]}"; do
+    if [[ ! -f "${OUTPUT_DIR}/google/protobuf/${OBJC_NAME}${EXT}" ]]; then
+      RUN_PROTOC=yes
+    fi
+  done
+done
+
+# If we haven't decided to run protoc because of a missing file, check to see if
+# an input has changed.
+if [[ "${RUN_PROTOC}" != "yes" ]] ; then
   # Find the newest input file (protos, compiler, and this script).
   # (these patterns catch some extra stuff, but better to over sample than
   # under)
@@ -56,7 +128,7 @@ else
   # Find the oldest output file.
   readonly OldestOutput=$(find \
         "${OUTPUT_DIR}" \
-        -type f -name "*pbobjc.[hm]" -print0 \
+        -type f -name "*.pbobjc.[hm]" -print0 \
         | xargs -0 stat -f "%m %N" \
         | sort -n -r | tail -n1 | cut -f2- -d" ")
   # If the newest input is newer than the oldest output, regenerate.
@@ -74,7 +146,7 @@ fi
 # Prune out all the files from previous generations to ensure we only have
 # current ones.
 find "${OUTPUT_DIR}" \
-    -type f -name "*pbobjc.[hm]" -print0 \
+    -type f -name "*.pbobjc.[hm]" -print0 \
     | xargs -0 rm -rf
 
 # -----------------------------------------------------------------------------
@@ -84,43 +156,12 @@ compile_protos() {
     --objc_out="${OUTPUT_DIR}/google/protobuf" \
     --proto_path=src/google/protobuf/          \
     --proto_path=src                           \
+    --experimental_allow_proto3_optional       \
     "$@"
 }
 
 # -----------------------------------------------------------------------------
-# Generate most of the proto files that exist in the C++ src tree.  Several
-# are used in the tests, but the extra don't hurt in that they ensure ObjC
-# sources can be generated from them.
-
-CORE_PROTO_FILES=(
-  src/google/protobuf/any_test.proto
-  src/google/protobuf/unittest_arena.proto
-  src/google/protobuf/unittest_custom_options.proto
-  src/google/protobuf/unittest_enormous_descriptor.proto
-  src/google/protobuf/unittest_embed_optimize_for.proto
-  src/google/protobuf/unittest_empty.proto
-  src/google/protobuf/unittest_import.proto
-  src/google/protobuf/unittest_import_lite.proto
-  src/google/protobuf/unittest_lite.proto
-  src/google/protobuf/unittest_mset.proto
-  src/google/protobuf/unittest_mset_wire_format.proto
-  src/google/protobuf/unittest_no_arena.proto
-  src/google/protobuf/unittest_no_arena_import.proto
-  src/google/protobuf/unittest_no_generic_services.proto
-  src/google/protobuf/unittest_optimize_for.proto
-  src/google/protobuf/unittest.proto
-  src/google/protobuf/unittest_import_public.proto
-  src/google/protobuf/unittest_import_public_lite.proto
-  src/google/protobuf/unittest_drop_unknown_fields.proto
-  src/google/protobuf/unittest_preserve_unknown_enum.proto
-  src/google/protobuf/map_lite_unittest.proto
-  src/google/protobuf/map_proto2_unittest.proto
-  src/google/protobuf/map_unittest.proto
-  # The unittest_custom_options.proto extends the messages in descriptor.proto
-  # so we build it in to test extending in general. The library doesn't provide
-  # a descriptor as it doesn't use the classes/enums.
-  src/google/protobuf/descriptor.proto
-)
+# Generate most of the proto files that exist in the C++ src tree.
 
 # Note: there is overlap in package.Message names between some of the test
 # files, so they can't be generated all at once. This works because the overlap
@@ -133,17 +174,4 @@ done
 # Generate the Objective C specific testing protos.
 compile_protos \
   --proto_path="objectivec/Tests" \
-  objectivec/Tests/unittest_cycle.proto \
-  objectivec/Tests/unittest_deprecated.proto \
-  objectivec/Tests/unittest_deprecated_file.proto \
-  objectivec/Tests/unittest_extension_chain_a.proto \
-  objectivec/Tests/unittest_extension_chain_b.proto \
-  objectivec/Tests/unittest_extension_chain_c.proto \
-  objectivec/Tests/unittest_extension_chain_d.proto \
-  objectivec/Tests/unittest_extension_chain_e.proto \
-  objectivec/Tests/unittest_extension_chain_f.proto \
-  objectivec/Tests/unittest_extension_chain_g.proto \
-  objectivec/Tests/unittest_runtime_proto2.proto \
-  objectivec/Tests/unittest_runtime_proto3.proto \
-  objectivec/Tests/unittest_objc.proto \
-  objectivec/Tests/unittest_objc_startup.proto
+  "${OBJC_TEST_PROTO_FILES[@]}"
